package colortable;
use strict;
use warnings;

our $use_percent;
my $default=__PACKAGE__->new();
sub BEGIN{ $use_percent=0; }


sub set_default
{
  my $self=shift;
  $default=$self if(ref($self) eq __PACKAGE__);
}

sub new
{
  my $class=shift;
  my $depth=shift || 8;
  my $pack=shift // 'C3';
  bless({
          update=>0,
          pack=>$pack,
          table=>[[0,0,0,pack($pack,0,0,0)]],
          bpc=>$depth,
          cmax=>(2**$depth)-1,
          cmin=>0,
          signals=>{
            changed=>[],
          },
          signal_lock=>0,
          signal_action=>{},
        },$class);
}

sub changed
{
  my $self=shift;
  $self=$default unless(ref($self));
  if(@_)
  {
    my $new=shift(@_)?1:0;
    $self->emitt_signal_changed if($new == 1 && $self->{update}!=$new);
    $self->{update}=$new;
  }
  return $self->{update};
}

sub table
{
  my $self=shift;
  $self=$default unless(ref($self));
  wantarray?@{$self->{table}}:$self->{table};
}

sub length
{
  my $self=shift;
  $self=$default unless(ref($self));
  scalar(@{$self->{table}});
}

sub add
{
  my $self=shift;
  $self=$default unless(ref($self));
  my $max=shift;
  my @color=@{$self->{table}->[-1]}[0..2];
  my @cup=(1,1,1);
  if($max)
  {
    my $i=$#{$self->table};
    while($i<=$max)
    {
      for my $p (0..2)
      {
        $cup[$p]*=-1 if(rand(423)>397);
        my $v=int($cup[$p]*rand(5));
        if($color[$p]+$v > $self->{cmax} || $color[$p]+$v < $self->{cmin})
        {
          $cup[$p]*=-1;
          $v*=-1;
        }
        $color[$p]+=$v;
      }
      $self->_append(@color);
      $self->_make_percent($i-1) if($use_percent);
      $i++;
    }
    $self->emitt_signal_changed();
  }
  else
  {
    my @tbl=$self->table;
    $tbl[0]->[3]=pack($self->{pack},@{$tbl[-1]}[0..2]);
    $tbl[-1]->[4]=[];
    my $i=0;
    while($i<=$#tbl)
    {
      $i++;
      my $entry=$tbl[$i-1];
      $entry->[3]=pack($self->{pack},@$entry[0..2]);
      $self->_make_percent($i-1) if($use_percent && !(ref($tbl[$i-1]->[4]) && @{$tbl[$i-1]->[4]}>=100));
    }
  }
}

sub _make_percent
{
  my $self=shift;
  $self=$default unless(ref($self));
  my $p=shift;
  my @ret;
  for(0..100)
  {
    my $c1=$self->table->[$p];
    my $c2=$self->table->[$p+1];
    my @l;
    for my $p (0..2)
    { push(@l,int($c1->[$p]+(($c2->[$p]-$c1->[$p])/100)*$_)); }
    push(@ret,pack($self->{pack},@l));
  }
  $self->table->[$p]->[4]=\@ret;
}

sub _append
{
  my $self=shift;
  $self=$default unless(ref($self));
  push(@{$self->{table}},[@_[0..2],pack($self->{pack},@_[0..2]),[]]);
}

sub clean
{
  my $self=shift;
  $self=$default unless(ref($self));
  @{$self->{table}}=(shift(@{$self->{table}}));
  $self->emitt_signal_changed();
}

sub find
{
  my $self=shift;
  $self=$default unless(ref($self));
  my $val=shift;
  my $perc=$val-int($val);
  $val=int($val);
  $self->add($val) if($val>$#{$self->table});
  return $self->table->[$val]->[4]->[100-int($perc*100)] if($use_percent);
  return $self->table->[$val]->[3];
}

sub save
{
  my $self=shift;
  $self=$default unless(ref($self));
  my $file=shift // return 0;
  return 0 if(@{$self->table()}<=1);
  open(my $fh,'>',$file) or return 0;
  my $v=($self->{cmax}-$self->{cmin});
  for($self->table())
  { print $fh join(';',map{$_/$v}@$_[0..2])."\n"; }
  close($fh) or return 0;
  return 1;
}

sub load
{
  my $self=shift;
  $self=$default unless(ref($self));
  my $file=shift // return 0;
  open(my $fh,'<',$file) or return 0;
  @{$self->table}=();
  my $v=($self->{cmax}-$self->{cmin});
  LOOP:
  while(<$fh>)
  {
    chomp;
    next if(/^\s*#/);
    s/\s*#.+$//;
    s/^\s+//;
    s/\s+$//;
    my @color=(split(/\s*;\s*/,$_))[0..2];
    next if(@color!=3);
    for(@color)
    {
      next LOOP if($_!~/^[\d\.e]$/i);
      $_*=$v;
      $_=$self->{cmax} if($_>$self->{cmax});
      $_=$self->{cmin} if($_<$self->{cmin});
    }
    push(@{$self->table},\@color);
  }
  close($fh);
  $self->add();
  $self->emitt_signal_changed();
  return 1;
}

sub array_to_bin
{
  my $self=shift;
  $self=$default unless(ref($self));

  my $array=shift;
  $array=[] unless($array || ref($array) eq 'ARRAY');

  my $call=shift;
  $call=undef if($call && ref($call) ne 'CODE');

  my $count=shift // 100;

  my $max=shift;

  my $data='';
  if($max)
  {
    $self->add($max+1) if($max > $#{$self->table});
    if($call)
    {
      my $max=$#$array;
      $count=int($#$array/$count);
      for my $pos (0..$max)
      {
        $call->($pos,$max) if($pos%$count==0);
        my $v=int($array->[$pos]);
        if($use_percent)
        {
          my $vv=int(($array->[$pos]-$v)*100);
          $data.=$self->{table}->[$v]->[4]->[100-$vv];
          next;
        }
        $data.=$self->{table}->[$v]->[3];
      }
    }
    else
    {
      for (@$array)
      {
        my $v=int($_);
        if($use_percent)
        {
          my $vv=int(($_-$v)*100);
          $data.=$self->{table}->[$v]->[4]->[100-$vv];
          next;
        }
        $data.=$self->{table}->[$v]->[3];
      }
    }
  }
  else
  {
    if($call)
    {
      my $max=$#$array;
      $count=int($#$array/$count);
      for my $pos (0..$max)
      {
        $call->($pos,$max) if($pos%$count==0);
        $data.=$self->find($array->[$pos]);
      }
    }
    else
    { $data.=$self->find($_) for (@$array); }
  }
  return $data;
}

sub example
{
  my $self=shift;
  $self=$default unless(ref($self));

  my $height=shift // 10;
  my $width=$self->length();
  $self->add();

  my $data='';
  $data.=$_->[3] for($self->table());

  return ($data x $height),$width,$height,$self->{bpc};
}

########################################################################
########################################################################
########################################################################

sub signal_connect
{
  my $self=shift;
  $self=$default unless(ref($self));
  my $name=shift // '';
  my $code=shift;
  my $opts=shift;
  return undef unless(ref($code) eq 'CODE');
  return undef unless(exists($self->{signals}->{$name}));
  push(@{$self->{signals}->{$name}},[$code,$opts]);
  return $name.@{$self->{signals}->{$name}};
}

sub signal_disconnect
{
  my $self=shift;
  $self=$default unless(ref($self));
  my $name=shift;
  return unless($name);
  return 0 unless($name=~/^(\w+?)(\d+)$/);
  return delete($self->{signals}->{$1}->[$2]);
}

sub emitt_signal_changed
{
  my $self=shift;
  $self=$default unless(ref($self));
  if($self->{signal_lock})
  { $self->{signal_action}->{'changed'}++; }
  else
  {
    for(@{$self->{signals}->{'changed'}})
    {
      next unless($_);
      $_->[0]->($self,$_->[1]);
    }
    $self->{update}=1;
  }
}

sub block_signals
{
  my $self=shift;
  $self=$default unless(ref($self));
  my $on=shift;
  return $self->{signal_lock}=1 if($on);
  $self->{signal_lock}=0;
  for(keys(%{$self->{signal_action}}))
  { $self->emitt_signal_changed() if($_ eq 'changed'); }
  $self->{signal_action}={};
  return 0;
}
1;
