package history;
use strict;
use warnings;

my $default=__PACKAGE__->new();

sub set_default
{
  my $self=shift;
  $default=$self if(ref($self) eq __PACKAGE__);
}

sub new
{
  my $class=shift;
  my $map=shift;

  $map=[] unless ($map && ref($map) eq  'ARRAY');
  my $cnt=0;
  $map={map{$_=>$cnt++}@$map};

  return bless({list=>[], pos=>-1, map=>$map},$class);
}

sub pos_is_last
{
  my $self=shift;
  $self=$default unless(ref($self));
  return $#{$self->{list}}<=$self->{pos}?1:0;
}

sub pos_is_first
{
  my $self=shift;
  $self=$default unless(ref($self));
  return $self->{pos}<=0?1:0;
}

sub _has_pos
{
  my $self=shift;
  $self=$default unless(ref($self));
  return 1 if(defined($_[0]) && $_[0]>=0 && @{$self->{list}}>$_[0]);
  return 0;
}

sub pos
{
  my $self=shift;
  $self=$default unless(ref($self));
  if($self->_has_pos(@_))
  { $self->{pos}=$_[0]; }
  return $self->{pos};
}

sub add
{
  my $self=shift;
  $self=$default unless(ref($self));
  if(@_)
  {
    $self->{pos}++;
    $self->_clean_up();
    $self->{list}->[$self->{pos}]=[@_];
  }
}

sub _clean_up
{
  my $self=shift;
  $self=$default unless(ref($self));
  pop(@{$self->{list}}) while($#{$self->{list}}>$self->{pos});
}

sub get
{
  my $self=shift;
  $self=$default unless(ref($self));
  my $pos=shift // $self->{pos};
  if($self->_has_pos($pos))
  {
    my $entry=$self->{list}->[$pos];
    if(@_)
    {
      my $key=shift;
      if(exists($self->{map}->{$key}))
      { return $self->{list}->[$pos]->[$self->{map}->{$key}]; }
      elsif($key=~/^\d+$/)
      { return $self->{list}->[$pos]->[$key]; }
      else
      { return undef; }
    }
    #print "$pos  => $self->{list}\n";
    return wantarray?@{$entry}:$entry;
  }
  return undef;
}

sub update
{
  my $self=shift;
  $self=$default unless(ref($self));
  my $pos=shift // $self->{pos};
  my $col=shift;
  my $replace=shift;
  if($self->_has_pos($pos) && defined($col))
  {
    my $entry=$self->{list}->[$pos];
    if(ref($col) eq 'ARRAY')
    {
      return 0 if(!defined($replace) || ref($replace) ne 'ARRAY');
      for(0..$#$col)
      {
        my $key=$col->[$_];
        my $val=$replace->[$_];
        if(my $pos = $self->{map}->{$key})
        { $key=$pos; }
        elsif($key!~/^\d+$/)
        { next; }
        $entry->[$key]=$val ;
      }
      return 1;
    }
    $entry->[$col]=$replace;
    return 1;
  }
  return 0;
}

sub all
{
  my $self=shift;
  $self=$default unless(ref($self));
  my $key=shift // '';
  if($key)
  {
    if(my $pos = $self->{map}->{$key})
    { $key=$pos; }
    elsif($key!~/^\d+$/)
    { return undef; }

    return map{$_->[$key]}@{$self->{list}};
  }
  return @{$self->{list}};
}

sub last
{
  my $self=shift;
  $self=$default unless(ref($self));
  $self->{pos}--;
  $self->{pos}=0 if($self->{pos}<0);
  return $self->get();
}

sub next
{
  my $self=shift;
  $self=$default unless(ref($self));
  $self->{pos}++;
  $self->{pos}=$#{$self->{list}} if($self->{pos}>$#{$self->{list}});
  return $self->get();
}

sub length
{
  my $self=shift;
  $self=$default unless(ref($self));
  return scalar(@{$self->{list}});
}

1;
